%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Toril Bye Rinnan
%   Engineering Cybernetics - June 2012 
%   
%   A nonlinear observer is implemented, along with
%   the developed EQUEST method and the EKF.
%   This files draws the Euler angles for the estimated 
%   attitude for all three methods and the sensor data
%   The file requires the function proj.m
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Local magnetic field for Trondheim
magnetometer=[13598.5;444.7;-49854.8];

%Normalized reference vector magnetometer
magnetometer=magnetometer/norm(magnetometer);
m1 = magnetometer(1);
m2 = magnetometer(2);
m3 = magnetometer(3);

%Normalized gravity vector
accelerometer=[0 0 1]';
a1=a(1);
a2=a(2);
a3=a(3);


%Initialization
qr = [0,0,0,1]';

x = zeros(10,1);
x(1:4,1) = [0 0 1 0]';
Ts = 0.0011;
go = true;
count=1;

q=[0,0,1,0]';
qq=[0,0,1,0]';
q_hat =[0,0,1,0]';
b_g = [0.0001 0.0001 0.0001]';
b_g_hat = [0,0,0]';

euler_ang1=[0,0,0]';
euler_ang2=[0,0,0]';
euler_ang3=[0,0,0]';
euler_ang4=[0,0,0]';

k = [1 1];
w = 0.1*rand(1);
simdata = zeros(2483,10);

%Covariance matrix for process and sensors
r1 = 0.05;
r2 = 0.08;
R = diag([r1,r1,r1,r2,r2,r2]);
P = 0*eye(10);

%Sampling time
dt = 0.01;
t=0;
N = 2484;

% %Estimate only for each 4. step
 for i=1:1:N-1
            
            %Gyroscope measurements
            omega = [0.1*sin(0.01*t) 0.1*sin(0.01*t) 0.1*sin(0.01*t)]';
            omega_real = omega + b_g;
            omega_real = omega_real*180/pi;
            
            q_vecr = [qr(1); qr(2); qr(3)];
            S=[0 -qr(3) qr(2);qr(3) 0 -qr(1);-qr(2) qr(1) 0];
            Rr = eye(3) + 2*qr(4)*S + 2*(S*S);
            
            %Normalized vector measurements
            mag = Rr'*magnetometer;
            acc = Rr'*accelerometer;            
            mag = mag/norm(mag);
            acc = acc/norm(acc);
                        
            %Attitude dynamics            
            T = 0.5*[-q_vecr'; (qr(4)*eye(3) + S)];
                
            qr = qr + dt*T*(omega_real-b_g);
            qr = qr/norm(qr);
            
            euler_ang1 = qua_to_euler(qr); 
            if t<0.04
                
                %EQUEST
            
                q_13=[0 -q(3) q(2);q(3) 0 -q(1);-q(2) q(1) 0];
                E=[q(4)*eye(3)+q_13;-q(1) -q(2) -q(3)];
                delta_q=0.5*E*omega_real;
                q=q+delta_q*0.01;
                q=q/norm(q);
    
                B=(1/0.001)*acc*[0 0 1] + (1/0.001)*mag*magnetometer';  %(1/0.05)*(mag/norm(mag))*magnetometer
                S=B+B';
                Z=[B(2,3)-B(3,2);B(3,1)-B(1,3);B(1,2)-B(2,1)];
                o=trace(B);
                V=[S-o*eye(3) Z;Z' o];
                %Q'*N*Q
                Q = [q(4) -q(1) -q(2) -q(3); q(1) q(4) -q(3) q(2); q(2) q(3) q(4) -q(1); q(3) -q(2) q(1) q(4)];
                N = Q'*eye(4)*Q*100;
                %KK = -V + N
                KK = -V*20+N;
            
                %Using eigenvalues to find optimal q   
                I = eye(4);
                [v,d] = eig(KK, I, 'chol');
                y = v(:,1);
                    
                %Converting from quaternions to euler angles(pitch,roll,yaw)
                q_hat=y(1:4);
                q_hat = q_hat/norm(q_hat);
                
            end
           
            
                %Observer gains
                k_i = 0.1;
                k = [0.5; 0.5];
                q_vec_hat = [q_hat(1); q_hat(2); q_hat(3)];
                S_hat = [0 -q_hat(3) q_hat(2);q_hat(3) 0 -q_hat(1);-q_hat(2) q_hat(1) 0];
                R_hat = eye(3) + 2*q_hat(4)*S_hat + 2*(S_hat*S_hat);
            
                v_hat_1= R_hat'*accelerometer;
                v_hat_2= R_hat'*magnetometer;
                v_1 = acc;
                v_2 = mag;
            
                %Injection term
                sum1 = 0.5*k(1)*((v_1*v_hat_1')-(v_hat_1*v_1'));
                sum2 = 0.5*k(2)*((v_2*v_hat_2')-(v_hat_2*v_2'));
                sum = sum1 + sum2;
            
                sigma = -[sum(3,2) sum(1,3) sum(2,1)]';
                        
                %Observer equations
           
                T_hat = 0.5*[-q_vec_hat'; q_hat(4)*eye(3) + S_hat];
                q_hat = q_hat + dt*T_hat*(omega_real - b_g_hat + sigma);         
                b_g_hat = proj(b_g_hat,(-k_i*sigma));
            
                q_hat = q_hat/norm(q_hat);
                euler_ang2 = qua_to_euler(q_hat);
            
                    
            %EKF
            z = [acc;mag];
            omega_skew=[0 -omega_real(3) omega_real(2);omega_real(3) 0 -omega_real(1);-omega_real(2) omega_real(1) 0];
            Omega = 0.5 * [omega_skew omega_real; -omega_real' 0];
    
            A = [expm(Omega*Ts) zeros(4,3) zeros(4,3);
            zeros(3,4) eye(3) zeros(3,3);
            zeros(3,4) zeros(3,3) eye(3)];
        
            %Extended Kalman Filter Equations
            
            %Time update eq
            x = A*x;
            x(1:4)=x(1:4)/norm(x(1:4));
    
            q1 = x(1);
            q2 = x(2);
            q3 = x(3);
            q4 = x(4);
            q_13=[0 -q3 q2;q3 0 -q1;-q2 q1 0];
            E=[q4*eye(3)+q_13;-q1 -q2 -q3];
            Y=[q4*eye(3)-q_13;-q1 -q2 -q3];
            Rot=E'*Y;
            
            %Linearization of measurment matrix
            num = (q1^2+q2^2+q3^2+q4^2)^(3/2);

            daxdq1 = a1*q1^3+3*a1*q1*q2^2+3*a1*q1*q3^2+a1*q1*q4^2-2*a2*q1*q3*q4+2*a2*q2^3+2*a2*q2*q3^2+2*a2*q2*q4^2+2*a3*q1*q2*q4+2*a3*q2^2*q3+2*a3*q3^3+2*a3*q3*q4^2;
            daxdq2 = -3*a1*q1^2*q2-a1*q2^3-a1*q2*q3^2-3*a1*q2*q4^2+2*a2*q1^3+2*a2*q1*q3^2+2*a2*q1*q4^2-2*a2*q2*q3*q4-2*a3*q1^2*q4-2*a3*q1*q2*q3-2*a3*q3^2*q4-2*a3*q4^3;
            daxdq3 = -3*a1*q1^2*q3-a1*q2^2*q3-a1*q3^3-3*a1*q3*q4^2+2*a2*q1^2*q4-2*a2*q1*q2*q3+2*a2*q2^2*q4+2*a2*q4^3+2*a3*q1^3+2*a3*q1*q2^2+2*a3*q1*q4^2+2*a3*q2*q3*q4;
            daxdq4 = -(-a1*q1^2*q4-3*a1*q2^2*q4-3*a1*q3^2*q4-a1*q4^3-2*a2*q1^2*q3+2*a2*q1*q2*q4-2*a2*q2^2*q3-2*a2*q3^3+2*a3*q1^2*q2+2*a3*q1*q3*q4+2*a3*q2^3+2*a3*q2*q3^2);
            daydq1 = 2*(a1*(q1*q3*q4+q2^3+q2*(q3^2+q4^2))+a3*(q4*(q2^2+q3^2+q4^2)-q1*q2*q3))-a2*q1*(q1^2+3*q2^2+q3^2+3*q4^2);
            daydq2 = 2*a1*q1^3+2*a1*q1*q3^2+2*a1*q1*q4^2+2*a1*q2*q3*q4+3*a2*q1^2*q2+a2*q2^3+3*a2*q2*q3^2+a2*q2*q4^2+2*a3*q1^2*q3-2*a3*q1*q2*q4+2*a3*q3^3+2*a3*q3*q4^2;
            daydq3 = -2*a1*q1^2*q4-2*a1*q1*q2*q3-2*a1*q2^2*q4-2*a1*q4^3-a2*q1^2*q3-3*a2*q2^2*q3-a2*q3^3-3*a2*q3*q4^2+2*a3*q1^2*q2-2*a3*q1*q3*q4+2*a3*q2^3+2*a3*q2*q4^2;
            daydq4 = -2*a1*q1^2*q3-2*a1*q1*q2*q4-2*a1*q2^2*q3-2*a1*q3^3+3*a2*q1^2*q4+2*a2*q2^2*q4+3*a2*q3^2*q4+a2*q4^3+2*a3*q1^3+2*a3*q1*q2^2+2*a3*q1*q3^2-2*a3*q2*q3*q4;
            dazdq1 = -(2*a1*q1*q2*q4-2*a1*q2^2*q3-2*a1*q3^3-2*a1*q3*q4^2+2*a2*q1*q2*q3+2*a2*q2^2*q4+2*a2*q3^2*q4+2*a2*q4^3+a3*q1^3+a3*q1*q2^2+3*a3*q1*q3^2+3*a3*q1*q4^2);
            dazdq2 = 2*(a1*(q1^2*q4-q1*q2*q3+q3^2*q4+q4^3)+a2*(q1^2*q3+q1*q2*q4+q3^3+q3*q4^2))-a3*q2*(q1^2+q2^2+3*(q3^2+q4^2));
            dazdq3 = 2*a1*q1^3+2*a1*q1*q2^2+2*a1*q1*q4^2-2*a1*q2*q3*q4+2*a2*q1^2*q2+2*a2*q1*q3*q4+2*a2*q2^3+2*a2*q2*q4^2+3*a3*q1^2*q3+3*a3*q2^2*q3+a3*q3^3+a3*q3*q4^2;
            dazdq4 = -(-2*a1*q1^2*q2+2*a1*q1*q3*q4-2*a1*q2^3-2*a1*q2*q3^2+2*a2*q1^3+2*a2*q1*q2^2+2*a2*q1*q3^2+2*a2*q2*q3*q4-3*a3*q1^2*q4-3*a3*q2^2*q4-a3*q3^2*q4-a3*q4^3);
    
            dmxdq1 = m1*q1^3+3*m1*q1*q2^2+3*m1*q1*q3^2+m1*q1*q4^2-2*m2*q1*q3*q4+2*m2*q2^3+2*m2*q2*q3^2+2*m2*q2*q4^2+2*m3*q1*q2*q4+2*m3*q2^2*q3+2*m3*q3^3+2*m3*q3*q4^2;
            dmxdq2 = -3*m1*q1^2*q2-m1*q2^3-m1*q2*q3^2-3*m1*q2*q4^2+2*m2*q1^3+2*m2*q1*q3^2+2*m2*q1*q4^2-2*m2*q2*q3*q4-2*m3*q1^2*q4-2*m3*q1*q2*q3-2*m3*q3^2*q4-2*m3*q4^3;
            dmxdq3 = -3*m1*q1^2*q3-m1*q2^2*q3-m1*q3^3-3*m1*q3*q4^2+2*m2*q1^2*q4-2*m2*q1*q2*q3+2*m2*q2^2*q4+2*m2*q4^3+2*m3*q1^3+2*m3*q1*q2^2+2*m3*q1*q4^2+2*m3*q2*q3*q4;
            dmxdq4 = -(-m1*q1^2*q4-3*m1*q2^2*q4-3*m1*q3^2*q4-m1*q4^3-2*m2*q1^2*q3+2*m2*q1*q2*q4-2*m2*q2^2*q3-2*m2*q3^3+2*m3*q1^2*q2+2*m3*q1*q3*q4+2*m3*q2^3+2*m3*q2*q3^2);
            dmydq1 = 2*(m1*(q1*q3*q4+q2^3+q2*(q3^2+q4^2))+m3*(q4*(q2^2+q3^2+q4^2)-q1*q2*q3))-m2*q1*(q1^2+3*q2^2+q3^2+3*q4^2);
            dmydq2 = 2*m1*q1^3+2*m1*q1*q3^2+2*m1*q1*q4^2+2*m1*q2*q3*q4+3*m2*q1^2*q2+m2*q2^3+3*m2*q2*q3^2+m2*q2*q4^2+2*m3*q1^2*q3-2*m3*q1*q2*q4+2*m3*q3^3+2*m3*q3*q4^2;
            dmydq3 = -2*m1*q1^2*q4-2*m1*q1*q2*q3-2*m1*q2^2*q4-2*m1*q4^3-m2*q1^2*q3-3*m2*q2^2*q3-m2*q3^3-3*m2*q3*q4^2+2*m3*q1^2*q2-2*m3*q1*q3*q4+2*m3*q2^3+2*m3*q2*q4^2;
            dmydq4 = -2*m1*q1^2*q3-2*m1*q1*q2*q4-2*m1*q2^2*q3-2*m1*q3^3+3*m2*q1^2*q4+2*m2*q2^2*q4+3*m2*q3^2*q4+m2*q4^3+2*m3*q1^3+2*m3*q1*q2^2+2*m3*q1*q3^2-2*m3*q2*q3*q4;
            dmzdq1 = -(2*m1*q1*q2*q4-2*m1*q2^2*q3-2*m1*q3^3-2*m1*q3*q4^2+2*m2*q1*q2*q3+2*m2*q2^2*q4+2*m2*q3^2*q4+2*m2*q4^3+m3*q1^3+m3*q1*q2^2+3*m3*q1*q3^2+3*m3*q1*q4^2);
            dmzdq2 = 2*(m1*(q1^2*q4-q1*q2*q3+q3^2*q4+q4^3)+m2*(q1^2*q3+q1*q2*q4+q3^3+q3*q4^2))-m3*q2*(q1^2+q2^2+3*(q3^2+q4^2));
            dmzdq3 = 2*m1*q1^3+2*m1*q1*q2^2+2*m1*q1*q4^2-2*m1*q2*q3*q4+2*m2*q1^2*q2+2*m2*q1*q3*q4+2*m2*q2^3+2*m2*q2*q4^2+3*m3*q1^2*q3+3*m3*q2^2*q3+m3*q3^3+m3*q3*q4^2;
            dmzdq4 = -(-2*m1*q1^2*q2+2*m1*q1*q3*q4-2*m1*q2^3-2*m1*q2*q3^2+2*m2*q1^3+2*m2*q1*q2^2+2*m2*q1*q3^2+2*m2*q2*q3*q4-3*m3*q1^2*q4-3*m3*q2^2*q4-m3*q3^2*q4-m3*q4^3);
    
    
            H = 1/num*[daxdq1 daxdq2 daxdq3 daxdq4 num 0 0 0 0 0;
               daydq1 daydq2 daydq3 daydq4 0 num 0 0 0 0;
               dazdq1 dazdq2 dazdq3 dazdq4 0 0 num 0 0 0;
               dmxdq1 dmxdq2 dmxdq3 dmxdq4 0 0 0 num 0 0;
               dmydq1 dmydq2 dmydq3 dmydq4 0 0 0 0 num 0;
               dmzdq1 dmzdq2 dmzdq3 dmzdq4 0 0 0 0 0 num];
           
            %(Ts/2)^2*E*10*eye(3)*E'
            Q = [0.001*eye(4) zeros(4,3) zeros(4,3);
            zeros(3,4) 0.00000001*eye(3) zeros(3,3);
            zeros(3,4) zeros(3,3) eye(3)*0.00000001];
    
            P = A*P*A'+Q;       
            %Measurement update eq
            %Iteratively updating of the estimated quaternion
            K = P*H'*inv(H*P*H'+R);
            x = x+K*(z-[Rot*[0 0 1]';Rot*magnetometer]-x(5:10));
            P = (eye(size(K*H))-K*H)*P;
    
    
            %Converting from quaternions to Euler angles(pitch,roll,yaw)
            qq = x(1:4)/norm(x(1:4));
            euler_ang3=qua_to_euler(qq);
            
            simdata(i,:) = [t euler_ang2'*180/pi euler_ang3'*180/pi euler_ang1'*180/pi];
            t=t+0.04;
            
 end

%Plot for Euler angles
 
t     = simdata(:,1); 
comb     = simdata(:,2:4); 
Kalman = simdata(:,5:7); 
q = simdata(:,8:10);

 
clf
figure(gcf)

subplot(311),plot(t,q(:,1),'g',t,comb(:,1),'r',t,Kalman(:,1),'b')
xlabel('Time (s)','Fontsize',14),ylabel('Roll','Fontsize',14),grid
h_legend = legend('Sensor data','Combination method','Nonlinear observer');
axis([0 inf -180 180]);
set(h_legend,'FontSize',12);


subplot(312),plot(t,q(:,2),'g',t,Equest(:,2),'r',t,Nonlinear(:,2),'k',t,Kalman(:,2),'b')
xlabel('Time (s)','Fontsize',14),ylabel('Pitch','Fontsize',14),grid
h_legend = legend('Sensor data','Combination method','Nonlinear observer');
axis([0 inf -180 180]);
set(h_legend,'FontSize',12);

subplot(313),plot(t,q(:,3),'g',t,Equest(:,3),'r',t,Nonlinear(:,3),'k',t,Kalman(:,3),'b')
xlabel('Time (s)','Fontsize',14),ylabel('Yaw','Fontsize',14),grid
h_legend = legend('Sensor data','Combination method','Nonlinear observer');
axis([0 inf -180 180]);
set(h_legend,'FontSize',12);
