%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Toril Bye Rinnan
%   Engineering Cybernetics - June 2012 
%   
%   This code is based on the implementation for the
%   original EQUEST method made by Yabar & Jenssen in April 2011.
%   
%   Further development of the algorithm is implemented, with the use of
%   multiplications instead of subtractions in the method's cost function.
%
%   This file runs the new developed EQUEST method and the EKF 
%   on saved data from the sparkfun IMU
%   "data_read2.m" is required to run the file.
%   This files draws the Euler angles for the estimated attitude 
%   both for the EKF and the new developed EQUEST method
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

data_read2

%Local magnetic field for Trondheim
magnetometer=[13598.5;444.7;-49854.8];

%Normalized reference vector magnetometer
magnetometer=magnetometer/norm(magnetometer);
m1 = magnetometer(1);
m2 = magnetometer(2);
m3 = magnetometer(3);

%Normalized gravity vector
a=[0 0 1];
a1=a(1);
a2=a(2);
a3=a(3);

%Initialization
q=[0,0,0,1]';
x = zeros(10,1);
x(1:4,1) = [0,0,1,0]';
qq=[0,0,1,0]';
P = 0*eye(10);
go = true;
count=1;
t=0;
euler_ang1=[0,0,0]';
euler_ang2=[0,0,0]';
Ts = 0.0011;

%Covariance matrix for process and sensors
r1 = 0.05;
r2 = 0.08;
R = diag([r1,r1,r1,r2,r2,r2]);

% %Estimate only for each 4. step
 for i=1:4:2483
            
            %Normalized vector measurements read from txt- file
            acc = dataM(3:5,i);
            acc = acc/norm(acc);
            
            mag = dataM(9:11,i);
            mag = mag/norm(mag);
            
            omega_real = dataM(6:8,i)-[260 257 260]';
            omega_real = omega_real*0.92*(pi/180);
                    
            %EKF
            z = [acc;mag];
            omega_skew=[0 -omega_real(3) omega_real(2);omega_real(3) 0 -omega_real(1);-omega_real(2) omega_real(1) 0];
            Omega = 0.5 * [omega_skew omega_real; -omega_real' 0];
                         
            A = [expm(Omega*Ts) zeros(4,3) zeros(4,3);
            zeros(3,4) eye(3) zeros(3,3);
            zeros(3,4) zeros(3,3) eye(3)];
        
            %Extended Kalman Filter Equations
            %Time update eq
            x = A*x;
            x(1:4)=x(1:4)/norm(x(1:4));
    
            q1 = x(1);
            q2 = x(2);
            q3 = x(3);
            q4 = x(4);
            q_13=[0 -q3 q2;q3 0 -q1;-q2 q1 0];
            E=[q4*eye(3)+q_13;-q1 -q2 -q3];
            Y=[q4*eye(3)-q_13;-q1 -q2 -q3];
            Rot=E'*Y;
                               
            %Linearization of measurment matrix
            num = (q1^2+q2^2+q3^2+q4^2)^(3/2);

            daxdq1 = a1*q1^3+3*a1*q1*q2^2+3*a1*q1*q3^2+a1*q1*q4^2-2*a2*q1*q3*q4+2*a2*q2^3+2*a2*q2*q3^2+2*a2*q2*q4^2+2*a3*q1*q2*q4+2*a3*q2^2*q3+2*a3*q3^3+2*a3*q3*q4^2;
            daxdq2 = -3*a1*q1^2*q2-a1*q2^3-a1*q2*q3^2-3*a1*q2*q4^2+2*a2*q1^3+2*a2*q1*q3^2+2*a2*q1*q4^2-2*a2*q2*q3*q4-2*a3*q1^2*q4-2*a3*q1*q2*q3-2*a3*q3^2*q4-2*a3*q4^3;
            daxdq3 = -3*a1*q1^2*q3-a1*q2^2*q3-a1*q3^3-3*a1*q3*q4^2+2*a2*q1^2*q4-2*a2*q1*q2*q3+2*a2*q2^2*q4+2*a2*q4^3+2*a3*q1^3+2*a3*q1*q2^2+2*a3*q1*q4^2+2*a3*q2*q3*q4;
            daxdq4 = -(-a1*q1^2*q4-3*a1*q2^2*q4-3*a1*q3^2*q4-a1*q4^3-2*a2*q1^2*q3+2*a2*q1*q2*q4-2*a2*q2^2*q3-2*a2*q3^3+2*a3*q1^2*q2+2*a3*q1*q3*q4+2*a3*q2^3+2*a3*q2*q3^2);
            daydq1 = 2*(a1*(q1*q3*q4+q2^3+q2*(q3^2+q4^2))+a3*(q4*(q2^2+q3^2+q4^2)-q1*q2*q3))-a2*q1*(q1^2+3*q2^2+q3^2+3*q4^2);
            daydq2 = 2*a1*q1^3+2*a1*q1*q3^2+2*a1*q1*q4^2+2*a1*q2*q3*q4+3*a2*q1^2*q2+a2*q2^3+3*a2*q2*q3^2+a2*q2*q4^2+2*a3*q1^2*q3-2*a3*q1*q2*q4+2*a3*q3^3+2*a3*q3*q4^2;
            daydq3 = -2*a1*q1^2*q4-2*a1*q1*q2*q3-2*a1*q2^2*q4-2*a1*q4^3-a2*q1^2*q3-3*a2*q2^2*q3-a2*q3^3-3*a2*q3*q4^2+2*a3*q1^2*q2-2*a3*q1*q3*q4+2*a3*q2^3+2*a3*q2*q4^2;
            daydq4 = -2*a1*q1^2*q3-2*a1*q1*q2*q4-2*a1*q2^2*q3-2*a1*q3^3+3*a2*q1^2*q4+2*a2*q2^2*q4+3*a2*q3^2*q4+a2*q4^3+2*a3*q1^3+2*a3*q1*q2^2+2*a3*q1*q3^2-2*a3*q2*q3*q4;
            dazdq1 = -(2*a1*q1*q2*q4-2*a1*q2^2*q3-2*a1*q3^3-2*a1*q3*q4^2+2*a2*q1*q2*q3+2*a2*q2^2*q4+2*a2*q3^2*q4+2*a2*q4^3+a3*q1^3+a3*q1*q2^2+3*a3*q1*q3^2+3*a3*q1*q4^2);
            dazdq2 = 2*(a1*(q1^2*q4-q1*q2*q3+q3^2*q4+q4^3)+a2*(q1^2*q3+q1*q2*q4+q3^3+q3*q4^2))-a3*q2*(q1^2+q2^2+3*(q3^2+q4^2));
            dazdq3 = 2*a1*q1^3+2*a1*q1*q2^2+2*a1*q1*q4^2-2*a1*q2*q3*q4+2*a2*q1^2*q2+2*a2*q1*q3*q4+2*a2*q2^3+2*a2*q2*q4^2+3*a3*q1^2*q3+3*a3*q2^2*q3+a3*q3^3+a3*q3*q4^2;
            dazdq4 = -(-2*a1*q1^2*q2+2*a1*q1*q3*q4-2*a1*q2^3-2*a1*q2*q3^2+2*a2*q1^3+2*a2*q1*q2^2+2*a2*q1*q3^2+2*a2*q2*q3*q4-3*a3*q1^2*q4-3*a3*q2^2*q4-a3*q3^2*q4-a3*q4^3);
    
            dmxdq1 = m1*q1^3+3*m1*q1*q2^2+3*m1*q1*q3^2+m1*q1*q4^2-2*m2*q1*q3*q4+2*m2*q2^3+2*m2*q2*q3^2+2*m2*q2*q4^2+2*m3*q1*q2*q4+2*m3*q2^2*q3+2*m3*q3^3+2*m3*q3*q4^2;
            dmxdq2 = -3*m1*q1^2*q2-m1*q2^3-m1*q2*q3^2-3*m1*q2*q4^2+2*m2*q1^3+2*m2*q1*q3^2+2*m2*q1*q4^2-2*m2*q2*q3*q4-2*m3*q1^2*q4-2*m3*q1*q2*q3-2*m3*q3^2*q4-2*m3*q4^3;
            dmxdq3 = -3*m1*q1^2*q3-m1*q2^2*q3-m1*q3^3-3*m1*q3*q4^2+2*m2*q1^2*q4-2*m2*q1*q2*q3+2*m2*q2^2*q4+2*m2*q4^3+2*m3*q1^3+2*m3*q1*q2^2+2*m3*q1*q4^2+2*m3*q2*q3*q4;
            dmxdq4 = -(-m1*q1^2*q4-3*m1*q2^2*q4-3*m1*q3^2*q4-m1*q4^3-2*m2*q1^2*q3+2*m2*q1*q2*q4-2*m2*q2^2*q3-2*m2*q3^3+2*m3*q1^2*q2+2*m3*q1*q3*q4+2*m3*q2^3+2*m3*q2*q3^2);
            dmydq1 = 2*(m1*(q1*q3*q4+q2^3+q2*(q3^2+q4^2))+m3*(q4*(q2^2+q3^2+q4^2)-q1*q2*q3))-m2*q1*(q1^2+3*q2^2+q3^2+3*q4^2);
            dmydq2 = 2*m1*q1^3+2*m1*q1*q3^2+2*m1*q1*q4^2+2*m1*q2*q3*q4+3*m2*q1^2*q2+m2*q2^3+3*m2*q2*q3^2+m2*q2*q4^2+2*m3*q1^2*q3-2*m3*q1*q2*q4+2*m3*q3^3+2*m3*q3*q4^2;
            dmydq3 = -2*m1*q1^2*q4-2*m1*q1*q2*q3-2*m1*q2^2*q4-2*m1*q4^3-m2*q1^2*q3-3*m2*q2^2*q3-m2*q3^3-3*m2*q3*q4^2+2*m3*q1^2*q2-2*m3*q1*q3*q4+2*m3*q2^3+2*m3*q2*q4^2;
            dmydq4 = -2*m1*q1^2*q3-2*m1*q1*q2*q4-2*m1*q2^2*q3-2*m1*q3^3+3*m2*q1^2*q4+2*m2*q2^2*q4+3*m2*q3^2*q4+m2*q4^3+2*m3*q1^3+2*m3*q1*q2^2+2*m3*q1*q3^2-2*m3*q2*q3*q4;
            dmzdq1 = -(2*m1*q1*q2*q4-2*m1*q2^2*q3-2*m1*q3^3-2*m1*q3*q4^2+2*m2*q1*q2*q3+2*m2*q2^2*q4+2*m2*q3^2*q4+2*m2*q4^3+m3*q1^3+m3*q1*q2^2+3*m3*q1*q3^2+3*m3*q1*q4^2);
            dmzdq2 = 2*(m1*(q1^2*q4-q1*q2*q3+q3^2*q4+q4^3)+m2*(q1^2*q3+q1*q2*q4+q3^3+q3*q4^2))-m3*q2*(q1^2+q2^2+3*(q3^2+q4^2));
            dmzdq3 = 2*m1*q1^3+2*m1*q1*q2^2+2*m1*q1*q4^2-2*m1*q2*q3*q4+2*m2*q1^2*q2+2*m2*q1*q3*q4+2*m2*q2^3+2*m2*q2*q4^2+3*m3*q1^2*q3+3*m3*q2^2*q3+m3*q3^3+m3*q3*q4^2;
            dmzdq4 = -(-2*m1*q1^2*q2+2*m1*q1*q3*q4-2*m1*q2^3-2*m1*q2*q3^2+2*m2*q1^3+2*m2*q1*q2^2+2*m2*q1*q3^2+2*m2*q2*q3*q4-3*m3*q1^2*q4-3*m3*q2^2*q4-m3*q3^2*q4-m3*q4^3);
    
    
            H = 1/num*[daxdq1 daxdq2 daxdq3 daxdq4 num 0 0 0 0 0;
               daydq1 daydq2 daydq3 daydq4 0 num 0 0 0 0;
               dazdq1 dazdq2 dazdq3 dazdq4 0 0 num 0 0 0;
               dmxdq1 dmxdq2 dmxdq3 dmxdq4 0 0 0 num 0 0;
               dmydq1 dmydq2 dmydq3 dmydq4 0 0 0 0 num 0;
               dmzdq1 dmzdq2 dmzdq3 dmzdq4 0 0 0 0 0 num];
           
            %(Ts/2)^2*E*10*eye(3)*E'
            Q = [0.001*eye(4) zeros(4,3) zeros(4,3);
            zeros(3,4) 0.00000001*eye(3) zeros(3,3);
            zeros(3,4) zeros(3,3) eye(3)*0.00000001];
    
            P = A*P*A'+Q;       
           
            %Measurement update eq
            %Iteratively updating of the estimated quaternion
            K = P*H'*inv(H*P*H'+R);
            x = x+K*(z-[Rot*[0 0 1]';Rot*magnetometer]-x(5:10));
            P = (eye(size(K*H))-K*H)*P;
    
    
            %Converting from quaternions to euler angles(pitch,roll,yaw)
            qq = x(1:4)/norm(x(1:4));
            euler_ang1=qua_to_euler(qq);
            
            %Developed EQUEST
            q_13=[0 -q(3) q(2);q(3) 0 -q(1);-q(2) q(1) 0];
            E=[q(4)*eye(3)+q_13;-q(1) -q(2) -q(3)];
            delta_q=0.5*E*omega_real;
            q=q+delta_q*0.01;
            q=q/norm(q);
    
            B=(1/0.001)*acc*[0 0 1] + (1/0.001)*mag*magnetometer';  %(1/0.05)*(mag/norm(mag))*magnetometer
            S=B+B';
            Z=[B(2,3)-B(3,2);B(3,1)-B(1,3);B(1,2)-B(2,1)];
            o=trace(B);
            V=[S-o*eye(3) Z;Z' o];
            %Q'*N*Q
            Q = [q(4) -q(1) -q(2) -q(3); q(1) q(4) -q(3) q(2); q(2) q(3) q(4) -q(1); q(3) -q(2) q(1) q(4)];
            N = Q'*eye(4)*Q*100;
            %KK = -V + N
            KK = -V+N;
            
            %Using eigenvalues to find optimal q   
            I = eye(4);
            [v,d] = eig(KK, I, 'chol');
            y = v(:,1);
                
            %Converting from quaternions to euler angles(pitch,roll,yaw)
            q=y(1:4);
            q = q/norm(q);
            euler_ang2 = qua_to_euler(q);
            
                        
            %Plot for Euler angles
            subplot(3,1,1)
            hold on;
            plot(t,euler_ang2(1)*180/pi,'color','red');
            plot(t,euler_ang1(1)*180/pi,'color','blue');
            axis([0 inf -180 180]); 
            h_legend = legend('dev. EQUEST','EKF');
            xlabel('Time [s]','Fontsize',14);
            ylabel('Roll [Deg]','Fontsize',14);
            set(h_legend,'FontSize',12);
            
            subplot(3,1,2)
            hold on;
            plot(t,euler_ang2(2)*180/pi,'color','red');
            plot(t,euler_ang1(2)*180/pi,'color','blue');
            axis([0 inf -180 180]); 
            xlabel('Time [s]','Fontsize',14);
            ylabel('Pitch [Deg]','Fontsize',14);
            
            subplot(3,1,3)
            hold on;
            plot(t,euler_ang2(3)*180/pi,'LineWidth',1,'color','red');
            plot(t,euler_ang1(3)*180/pi,'LineWidth',1,'color','blue');
            axis([0 inf -180 180]); 
            xlabel('Time [s]','Fontsize',14);
            ylabel('Yaw [Deg]','Fontsize',14);
            
            %Time update
            t=t+0.04;
            
 end